from __future__ import annotations

import sys
from enum import Enum
from typing import (
    List,
    Tuple,
    Union,
    Dict,
    Optional,
    Literal,
    Callable,
    TYPE_CHECKING,
    TypeVar,
    Any,
    Iterator,
    Iterable,
    AsyncIterator,
    MutableMapping,
    Sequence,
    Generator,
    AsyncGenerator,
    TypedDict,
    Coroutine,
    Type,
    Generic,
    Annotated,
    cast,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

Message = Dict[str, Any]
PyMode = Literal["REPL", "SCRIPT"]
Param = Union[bytes, str, float, bool, None, Sequence[Any], MutableMapping[str, Any]]


class JsonRpcCode(int, Enum):
    ParseError = -32700
    InvalidRequest = -32600
    MethodNotFound = -32601
    InvalidParams = -32602
    InternalError = -32603
    ServerError = -32000
    StopAsyncIterationError = -32099

Code = Union[int, JsonRpcCode]



import inspect
from typing import Any, Dict, List
import textwrap


def schema_gen(name: str, environment: Dict[str, Any], mainname: str = "") -> str:
    result = f"""
import typing

class {name}:
""".strip()

    for key, value in environment.items():
        if inspect.isfunction(value):
            signature = inspect.signature(value)
            result += (
                "\n"
                + textwrap.indent(
                    schema_gen_function(
                        mainname if mainname else name,
                        key,
                        signature,
                        value.__doc__ or "",
                        mainname,
                        value,
                    ),
                    " " * 4,
                )
                + "\n"
            )
        elif inspect.isclass(value):
            main_name = f"{mainname if mainname else name}.{key}"
            res = schema_gen(key, obj_to_dict(value), main_name).split("\n")[1:]
            res[
                2
            ] = f"    def __await__(self) -> typing.Generator[typing.Any, typing.Any, {main_name}]: ..."
            result += "\n    ".join(res)
        else:
            result += f"\n    {key}: typing.Awaitable[{type(value).__name__}]\n"

    return result


def schema_gen_function(
    base: str,
    name: str,
    signature: inspect.Signature,
    doc: str,
    mainname: str,
    value: Any,
) -> str:
    return_type = format_type(signature.return_annotation, value)
    args = list(map(str, signature.parameters.values()))
    args = ", ".join(args) if base != mainname else ", ".join(args[1:])
    result = f"""
class _Proxy_{name}(typing.Protocol):
    async def __call__(self, {args}) -> {return_type}: ...
""".strip()
    params: List[str] = []
    for arg, param in signature.parameters.items():
        if param.kind == inspect.Parameter.POSITIONAL_ONLY:
            raise ValueError(f"{arg} is positional only")
        params.append(
            f"def {arg}(self, value: {format_type(param.annotation, value)}, /) -> {base}._Proxy_{name}_collector: ..."
        )

    parmas_str = "\n".join(params)
    result += "\n" + textwrap.indent(parmas_str, prefix=" " * 4) + "\n"

    result += f"""
class _Proxy_{name}_collector(typing.Protocol):
    def __await__(self) -> typing.Generator[typing.Any, typing.Any, {return_type}]: ...
""".strip()

    result += "\n" + textwrap.indent(parmas_str, prefix=" " * 4) + "\n"

    result += f"{name}: _Proxy_{name}"
    result += f'\n"""{doc}"""'

    return result


def format_type(type_: Any, value: Any) -> str:
    if type_ is getattr(inspect, "_empty"):
        return "typing.AsyncGenerator" if inspect.isasyncgenfunction(value) else "typing.Any"
    if isinstance(type_, type):
        return type_.__name__
    return str(type_)


def obj_to_dict(obj) -> Dict[str, Any]:
    return {
        k: v
        for k, v in obj.__dict__.items()
        if not k in ("__dict__", "__weakref__", "__doc__", "__annotations__")
    }



from abc import ABC
from enum import Enum

from ..typing import Dict, Callable, Any, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from ..network.client.aioclient import AioClient, BaseClient
    from ..network.server.aioserver import AioServer, BaseServer

    Server = Union[AioServer, BaseServer]
    Client = Union[AioClient, BaseClient]


class JsonRpcMode(str, Enum):
    JsonRpc2 = "2.0"
    JsonRpcX = "X"
    Auto = "Auto"


class ListenEvent(str, Enum):
    Exception = "Exception"


class BaseJsonRpc(ABC):
    _server: "Server"
    _client: "Client"
    _PATH: str
    _HOST: str
    _PORT: int
    _Execer: bool = False
    _JsonRpcMode: JsonRpcMode = JsonRpcMode.Auto
    _Events: Dict[ListenEvent, Callable] = {}
    _namespace: Dict[str, Any] = {}

    def on(self, event: ListenEvent, callback) -> None:
        self._Events[event] = callback

    async def __aenter__(self) -> "BaseJsonRpc":
        return self

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        ...



import asyncio
import inspect
from functools import lru_cache

from pydantic import BaseModel

from ..exception import JsonRpcException
from ..typing import (
    Union,
    Tuple,
    List,
    Iterator,
    Dict,
    Optional,
    Iterable,
    Param,
    Code,
    JsonRpcCode,
    TYPE_CHECKING,
    TypeVar,
    Generic,
    Any,
)

if TYPE_CHECKING:
    from ..core.core import JsonRpc
    from .jsonrpc2 import JsonRpc2Request, JsonRpc2Response
    from .jsonrpcX import JsonRpcXRequest, JsonRpcXResponse


T_Res = TypeVar("T_Res", "JsonRpcXResponse", "JsonRpc2Response")
T_Req = TypeVar("T_Req", "JsonRpcXRequest", "JsonRpc2Request")


class JsonRpcError(BaseModel):
    code: Code
    message: str
    data: Optional[Param] = None


class Assign(Generic[T_Res, T_Req]):  # TODO 与其他Json-RPC实现进行交互时可能存在ID冲突

    future: Dict[str, asyncio.Event] = {}
    result: Dict[str, Any] = {}

    @classmethod
    async def receiver(cls, uid: str, pop: bool = True):
        event = asyncio.Event()
        cls.future[uid] = event
        await event.wait()
        if pop:
            data = cls.result.pop(uid)
        else:
            data = cls.result[uid]
            return data
        if "error" in data.dict(exclude_unset=True) and data.error != None:
            if data.error.code == JsonRpcCode.StopAsyncIterationError:
                cls.future.pop(uid, None)
                return JsonRpcCode.StopAsyncIterationError
            raise JsonRpcException(data.error.code, data.error.message, data.error.data)
        elif "result" in data.dict(exclude_unset=True) and data.result != None:
            return data.result
        raise TypeError("JsonRpc Response Obj Error")

    @classmethod
    def post_data(cls, data: Any, pop: bool = True):
        uid = data.id
        cls.future[uid].set()
        if pop:
            cls.future.pop(uid, None)
        cls.result[uid] = data


class ProxyObj:
    def __init__(self, _self: "JsonRpc"):
        self.__self = _self
        self.__methods: List[Union[str, Tuple[tuple, dict]]] = []

    def __getattr__(self, name) -> "ProxyObj":
        self.__methods.append(name)
        return self

    def __call__(self, *args, **kwargs) -> "ProxyObj":
        self.__methods.append((args, kwargs))
        return self

    def __iter__(self) -> Iterator[Union[str, Tuple[tuple, dict]]]:
        return iter(self.__methods)

    def __aiter__(self):
        self.__methods.append("__aiter__")
        self.__aiter_request = False
        return self

    async def __anext__(self):
        if hasattr(self.__self, "_client"):
            if not self.__self._client.closed():
                if not self.__aiter_request:
                    request = self.__self.auto_standard(self)
                    self.__aiter_request = request
                else:
                    request = self.__aiter_request
                if TYPE_CHECKING:
                    assert isinstance(request, (JsonRpcXRequest, JsonRpc2Request))
                try:
                    await self.__self._client.send_json(request.dict())
                    data = await Assign.receiver(request.id)
                except JsonRpcException as e:
                    raise e
                if (
                    isinstance(data, JsonRpcCode)
                    and data == JsonRpcCode.StopAsyncIterationError
                ):
                    raise StopAsyncIteration
                return data
        if not self.__aiter_request:
            request = self.__self.auto_standard(self)
            self.__aiter_request = request
        else:
            request = self.__aiter_request
        if TYPE_CHECKING:
            assert isinstance(request, (JsonRpcXRequest, JsonRpc2Request))
        try:
            frame = inspect.currentframe()
            ws = self.__self.get_ws(frame)
            await self.__self._server.send_json(ws, request.dict())
            data = await Assign.receiver(request.id)
        except JsonRpcException as e:
            raise e
        if (
            isinstance(data, JsonRpcCode)
            and data == JsonRpcCode.StopAsyncIterationError
        ):
            raise StopAsyncIteration
        return data

    def __await__(self):
        if hasattr(self.__self, "_client"):
            if not self.__self._client.closed():
                request = self.__self.auto_standard(self)
                yield from self.__self._client.send_json(request.dict()).__await__()
                try:
                    data = yield from Assign.receiver(request.id).__await__()
                except JsonRpcException as e:
                    raise e
                return data
        frame = inspect.currentframe()
        ws = self.__self.get_ws(frame)
        data = yield from self.__self.send_request_server(self, ws).__await__()
        return data


class Proxy:
    def __getattr__(self, name) -> ProxyObj:
        proxy = getattr(ProxyObj(self), name)  # type: ignore
        return proxy

    @classmethod
    @lru_cache(1024)
    def get_ws(cls, frame):
        for local in cls.__frame_locals(frame):
            for k,v in local.items():
                if 'websocket' == k: # TODO: 重构
                    return v
                # if 'exce_rpc' in k and 'websocket' in local:
                #     return local.get('websocket')
        raise AttributeError("websocket is not found")

    @staticmethod
    def __frame_locals(frame):
        frame = frame.f_back  # type: ignore
        while frame:
            yield frame.f_locals
            frame = frame.f_back


def analysis_proxy(
    proxy: Union[ProxyObj, Iterable[Union[str, Tuple[tuple, dict]]]]
) -> Iterator:
    for i in proxy:
        if isinstance(i, str):
            yield {"method": "__getattr__", "args": (i,), "kwargs": {}}
        else:
            yield {"method": "__call__", "args": i[0], "kwargs": i[1]}



import asyncio
import uuid
from json.decoder import JSONDecodeError
from pathlib import Path

from pydantic.error_wrappers import ValidationError

from .abcbase import BaseJsonRpc, JsonRpcMode, ListenEvent
from ..standard import (
    Proxy,
    ProxyObj,
    Assign,
    proxy_to_jsonrpc2,
    proxy_to_jsonrpcx,
    analysis_jsonrpc2_message,
    analysis_jsonrpcX_message,
    Standard,
    JsonRpc2Request,
    JsonRpc2Response,
    JsonRpcXRequest,
    JsonRpcXResponse,
    JsonRpcError,
    JsonRpcCode,
)

from ..state import State
from ..logs import logger
from ..exception import WebSocketClosedError
from ..schema import schema_gen
from ..utils import run_fm, asyncio_loop_apply, LaciaGenerator
from ..typing import (
    Any,
    Dict,
    Optional,
    Union,
    Param,
    TYPE_CHECKING,
    TypeVar,
    cast,
    Type,
    Callable,
)

if TYPE_CHECKING:
    from ..network.client.aioclient import AioClient, BaseClient
    from ..network.server.aioserver import AioServer, BaseServer
    from ..core.core import JsonRpc

    Server = Union[AioServer, BaseServer]
    Client = Union[AioClient, BaseClient]

Schema = TypeVar("Schema")


class JsonRpc(BaseJsonRpc, Proxy):  # TODO： Server 和 Client 分离
    def __init__(
        self,
        path: str = "",
        host: str = "localhost",
        port: int = 8080,
        execer: bool = True,
        jsonrpc_mode: JsonRpcMode = JsonRpcMode.Auto,
        namespace: Dict[str, Any] = {},
        loop: Optional[asyncio.AbstractEventLoop] = None,
        debug: bool = False,
    ):
        self._PATH = path
        self._HOST = host
        self._PORT = port
        self._Execer = execer
        self._JsonRpcMode = jsonrpc_mode
        self._namespace = namespace
        self._loop = loop or asyncio.get_event_loop()
        asyncio_loop_apply(self._loop)  # TODO: 不兼容审查
        State.debug = debug

    def run_server(self, server: "Server"):
        self._server = server
        self._server.set_on_ws(self.__listening_client, self)
        server.start(self._PATH, self._HOST, self._PORT, self._loop)

    async def run_client(self, client: "Client"):
        self._client = client
        await client.start(path=f"{self._HOST}:{self._PORT}{self._PATH}")
        self._loop.create_task(self.__listen_server())

    async def __listening_client(self, websocket) -> None:
        logger.info("🛰️ listening client")
        try:
            async for message in self._server.iter_json(websocket):
                try:
                    message_type = self.auto_request_response(message)
                    if message_type == "request":
                        msg = Standard.json_to_request(message)
                        if msg.id in Assign.future:
                            Assign.post_data(msg, False)
                        else:
                            self._loop.create_task(self.__exce_rpc(msg, websocket))
                    elif message_type == "response":
                        msg = Standard.json_to_response(message)
                        Assign.post_data(msg)  # type: ignore
                except ValidationError as e:
                    await self.__send_error_response(
                        websocket, JsonRpcCode.InvalidRequest, str(e), message
                    )
        except JSONDecodeError as e:
            logger.exception(e)
            await self.__send_error_response(websocket, JsonRpcCode.ParseError, str(e))
        except WebSocketClosedError as e:
            logger.debug(str(e))
        except Exception as e:
            logger.exception(e)
            f = self._Events.get(ListenEvent.Exception, None)
            if f:
                self._loop.create_task(f(e))

    async def __listen_server(self) -> None:
        logger.info("🛰️ listening server")
        try:
            async for message in self._client.iter_json():
                try:
                    message_type = self.auto_request_response(message)
                    if message_type == "request":
                        msg = Standard.json_to_request(message)
                        if msg.id in Assign.future:
                            Assign.post_data(msg, False)
                        else:
                            self._loop.create_task(self.__exce_rpc(msg, self._client))
                    elif message_type == "response":
                        msg = Standard.json_to_response(message)
                        Assign.post_data(msg)  # type: ignore
                except ValidationError as e:
                    await self.__send_error_response(
                        self._client, JsonRpcCode.InvalidRequest, str(e), message
                    )
        except JSONDecodeError as e:
            logger.exception(e)
            await self.__send_error_response(self.ws, JsonRpcCode.ParseError, str(e))
        except WebSocketClosedError as e:
            logger.debug(str(e))
        except Exception as e:
            logger.exception(e)
            f = self._Events.get(ListenEvent.Exception, None)
            if f:
                self._loop.create_task(f(e))

    async def __exce_rpc(
        self, message: Union[JsonRpc2Request, JsonRpcXRequest], websocket
    ) -> Any:
        logger.info(f"⚙️ try exce_rpc: {message}")
        try:
            if not self.execer:
                raise AttributeError("execer is not set correctly")
            if isinstance(message, JsonRpc2Request):
                fm = list(analysis_jsonrpc2_message(message))
            elif isinstance(message, JsonRpcXRequest):
                fm = list(analysis_jsonrpcX_message(message))
            else:
                raise TypeError("message is not correctly")
            result = await run_fm(fm, self._namespace)
            if isinstance(result, LaciaGenerator):
                await self.__handle_aiter(result, websocket, message)
            else:
                await self.__send_result(websocket, message, result)
        except AttributeError as e:
            logger.exception(e)
            await self.__send_error_response(
                websocket, JsonRpcCode.MethodNotFound, str(e), message.dict()
            )
        except TypeError as e:
            logger.exception(e)
            await self.__send_error_response(
                websocket, JsonRpcCode.ParseError, str(e), message.dict()
            )
        except Exception as e:
            logger.exception(e)
            await self.__send_error_response(
                websocket, JsonRpcCode.InternalError, str(e), message.dict()
            )

    async def __send_result(
        self, websocket, message: Union[JsonRpc2Request, JsonRpcXRequest], result: Param
    ) -> None:
        if isinstance(message, JsonRpc2Request):
            if self.__dict__.get('_server', None):
                await self._server.send_json(
                    websocket,
                    JsonRpc2Response(id=message.id, result=result).dict(exclude={"error"})
                )
            else:
                await self._client.send_json(
                    JsonRpc2Response(id=message.id, result=result).dict(exclude={"error"})
                )
        elif isinstance(message, JsonRpcXRequest):
            if self.__dict__.get('_server', None):
                await self._server.send_json(
                    websocket,
                    JsonRpcXResponse(id=message.id, result=result).dict(exclude={"error"})
                )
            else:
                await self._client.send_json(
                    JsonRpcXResponse(id=message.id, result=result).dict(exclude={"error"})
                )
        else:
            raise TypeError("message is not correctly")

    async def __send_request(self, proxy: ProxyObj):
        if hasattr(self, "_client"):
            if not self._client.closed():
                request = self.auto_standard(proxy)
                await self._client.send_json(request.dict())
                return await Assign.receiver(request.id)
        raise AttributeError("client is not running")

    async def __send_request_client(self, proxy: ProxyObj):
        request = self.auto_standard(proxy)
        await self._client.send_json(request.dict())
        return await Assign.receiver(request.id)

    async def send_request_server(self, proxy: ProxyObj, websocket):
        request = self.auto_standard(proxy)
        await self._server.send_json(websocket, request.dict())
        return await Assign.receiver(request.id)

    async def __send_error_response(
        self, websocket, error: JsonRpcCode, error_data: str, message: dict = {}
    ) -> None:
        logger.error(f"🚨 error: {error} {error._name_} {message}")
        if self._JsonRpcMode in (JsonRpcMode.Auto, JsonRpcMode.JsonRpc2):
            response = JsonRpc2Response(
                id=message.get("id", None),
                error=JsonRpcError(code=error, message=error._name_, data=error_data),
            ).dict(exclude={"result"})
        elif self._JsonRpcMode in (JsonRpcMode.Auto, JsonRpcMode.JsonRpcX):
            response = JsonRpcXResponse(
                id=message.get("id", None),
                error=JsonRpcError(code=error, message=error._name_, data=error_data),
            ).dict(exclude={"result"})
        else:
            raise ValueError("JsonRpcMode is not set correctly")
        if self.__dict__.get('_server', None):
            await self._server.send_json(websocket, response)
        else:
            await self._client.send_json(response)

    async def __handle_aiter(
        self, result: LaciaGenerator, websocket, message: Union[JsonRpc2Request, JsonRpcXRequest]
    ):  
        if hasattr(result.obj, "__aiter__"):
            async for r in result.obj:
                await self.__send_result(websocket, message, r)
                await Assign.receiver(message.id, False)
            await self.__send_error_response(
                websocket, JsonRpcCode.StopAsyncIterationError, "", message.dict()
            )
        elif hasattr(result.obj, "__iter__"):
            for r in result.obj:
                await self.__send_result(websocket, message, r)
                await Assign.receiver(message.id, False)
            await self.__send_error_response(
                websocket, JsonRpcCode.StopAsyncIterationError, "", message.dict()
            )
        else:
            raise TypeError("result is not iterable")

    def auto_standard(self, proxy: ProxyObj):  # TODO: 重构
        try:
            method, params = proxy_to_jsonrpc2(proxy)
        except TypeError:
            method = proxy_to_jsonrpcx(proxy)
        uid = str(uuid.uuid1())
        loc = locals()
        if self._JsonRpcMode == JsonRpcMode.Auto:
            return (
                JsonRpc2Request(id=uid, method=method, params=params)  # type: ignore
                if "params" in loc
                else JsonRpcXRequest(id=uid, method=method)  # type: ignore
            )
        elif self._JsonRpcMode == JsonRpcMode.JsonRpc2:
            if "params" in loc:
                return JsonRpc2Request(id=uid, method=method, params=params)  # type: ignore
        elif self._JsonRpcMode == JsonRpcMode.JsonRpcX:
            if "params" in loc:
                return JsonRpc2Request(id=uid, method=method, params=params)  # type: ignore
            else:
                return JsonRpcXRequest(id=uid, method=method)  # type: ignore
        raise ValueError("JsonRpcMode is not set correctly")

    def auto_request_response(self, message: dict):
        if "method" in message:
            return "request"
        if "result" in message or "error" in message:
            return "response"
        raise JSONDecodeError("message is not correctly", "", 0)

    async def __aenter__(self) -> "JsonRpc":
        ...

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        ...

    def add_namespace(self, name: str, v: Any):
        self._namespace[name] = v

    def with_schema(self, schema: Type[Schema]) -> Schema:
        return cast(schema, self)

    def generate_pyi(self, name: str, path: Optional[str] = None):
        if path == None:
            p = Path.cwd() / "laciaschema"
        else:
            p = Path(path)
        p.mkdir(parents=True, exist_ok=True)
        with (p / f"{name}.pyi").open("w") as f:
            f.write(schema_gen(name, self._namespace))
        with (p / f"{name}.py").open("w") as f:
            f.write(f"class {name}: ...\n")
        logger.success(f"{name}.pyi and {name}.py generated in {p}")



import sys
import atexit
import asyncio
import inspect
from pathlib import Path

from nest_asyncio import apply
from pydantic import BaseModel

from .typing import PyMode, Optional, Any, List, Callable, Dict


class CallObj(BaseModel):
    method: Callable
    args: tuple
    kwargs: dict

    def add_args(self, *args):
        self.args = (*self.args, *args)
        return self

    def add_kwargs(self, **kwargs):
        self.kwargs = {**self.kwargs, **kwargs}
        return self

    async def __call__(self):
        return await self.method(*self.args, **self.kwargs)


class LaciaGenerator:
    def __init__(self, obj) -> None:
        self.obj = obj


def get_func_bind(func: Callable, *args, **kwargs) -> Dict[str, Any]:
    """
    Get the function binding.
    """
    sig = inspect.signature(func)
    return {k: v for k, v in sig.bind(*args, **kwargs).arguments.items()}


def get_python_mode() -> PyMode:  # TODO: There is a bug when the main file and the interpreter working file are in the same directory
    """
    Get the python mode.
    """
    path = Path(sys.argv[0])
    if path.is_file():
        if Path(path.parent) == Path(sys.path[0]):
            return "SCRIPT"
        else:
            return "REPL"
    else:
        return "REPL"


def asyncio_loop_apply(loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
    return apply(loop)


async def run_obj(obj: Any, method: str, args: tuple = (), kwargs: dict = {}) -> Any:

    if method == "__getattr__":
        if args[0] == "__aiter__":
            return obj
        else:
            obj = getattr(obj, args[0])
        return obj
    elif method == "__call__":
        if asyncio.iscoroutinefunction(obj):
            res = await obj(*args, **kwargs)
        else:
            res = obj(*args, **kwargs)
        return res
    raise AttributeError("Method not found")


async def run_fm(fm: List[dict], namespace: dict):
    method = fm[0]
    if (
        method["method"] == "__getattr__"
        and method["args"][0] in namespace
        and (not method["kwargs"])
    ):
        obj = namespace[method["args"][0]]
    else:
        raise AttributeError("Method not found")
    for i in fm[1:]:
        obj = await run_obj(obj, i["method"], i["args"], i["kwargs"])
    if "i" in locals() and i["method"] == "__getattr__" and i["args"][0] == "__aiter__":  # type: ignore
        return LaciaGenerator(obj)
    return obj


def auto_close_loop():
    """
    Close the loop when the program exits.
    """
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        return
    try:
        asyncio.runners._cancel_all_tasks(loop)  # type: ignore
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.run_until_complete(loop.shutdown_default_executor())
    finally:
        loop.close()


atexit.register(auto_close_loop)



import asyncio
from abc import ABC, abstractmethod

from ..hook import Hook
from ..typing import Message, Optional, Tuple, TypeVar, List, Generic, Generator

W = TypeVar('W')

class BaseServer(Hook, Generic[W]):
    active_connections: List[Tuple[W, asyncio.Event]]

    @abstractmethod
    async def receive(self, websocket: W) -> Message:
        ...

    async def send(self, message) -> None:
        ...

    @abstractmethod
    async def receive_json(self, websocket: W) -> Message:
        ...

    @abstractmethod
    async def receive_bytes(self, websocket: W) -> bytes:
        ...

    @abstractmethod
    async def iter_bytes(self, websocket: W) -> Generator[bytes, None, None]:
        ...

    @abstractmethod
    async def iter_json(self, websocket: W) -> Generator[Message, None, None]:
        ...

    @abstractmethod
    async def send_bytes(self, websocket: W, message: bytes) -> None:
        ...

    @abstractmethod
    async def send_json(self, websocket: W, message: Message, binary = True) -> None:
        ...

    @abstractmethod
    async def start(self, *args, **kwargs) -> "BaseServer":
        ...

    @abstractmethod
    async def close(self, code: int, reason: Optional[str] = None) -> None:
        ...

    def closed(self) -> bool:
        ...

    @abstractmethod
    def set_on_ws(self, Callable, *args, **kwargs):
        ...

class BaseClient(Hook):
    @abstractmethod
    async def receive(self) -> Message:
        ...

    @abstractmethod
    async def send(self, message) -> None:
        ...

    @abstractmethod
    async def receive_json(self) -> Message:
        ...

    @abstractmethod
    async def receive_bytes(self) -> bytes:
        ...

    @abstractmethod
    async def iter_bytes(self) -> Generator[bytes, None, None]:
        ...

    @abstractmethod
    async def iter_json(self) -> Generator[Message, None, None]:
        ...

    @abstractmethod
    async def send_bytes(self, message: bytes) -> None:
        ...

    @abstractmethod
    async def send_json(self, message: Message, binary = True) -> None:
        ...

    @abstractmethod
    async def start(self, *args, **kwargs) -> "BaseClient":
        ...

    @abstractmethod
    async def close(self, code: int, reason: Optional[str] = None) -> None:
        ...

    @abstractmethod
    def closed(self) -> bool:
        ...


import aiohttp

import orjson
import bson

from ..abcbase import BaseClient
from ...logs import logger
from ...exception import WebSocketClosedError
from ...typing import Message, Optional, AsyncGenerator


class AioClient(BaseClient):
    def __init__(self) -> None:
        self.session = aiohttp.ClientSession()

    async def start(self, path: str) -> "AioClient":
        self.ws = await self.session.ws_connect(f"http://{path}")
        logger.success(f"📡 {self.__class__.__name__} success connected: {path}.")
        return self

    async def receive(self):
        try:
            async for data in self.ws:
                if data.type == aiohttp.WSMsgType.close:
                    raise WebSocketClosedError(data.data)
                else:
                    return data
        finally:
            if not 'data' in locals():
                await self.close()
                raise WebSocketClosedError(f"{self.__class__.__name__} closed.")

    async def receive_json(self):
        data = await self.receive()
        if data and data.type == aiohttp.WSMsgType.TEXT:
            data = orjson.loads(data.data)
            return data
        elif data and data.type == aiohttp.WSMsgType.BINARY:
            data = bson.loads(data.data)
            return data

    async def receive_bytes(self):
        data = await self.receive()
        if data and data.type == aiohttp.WSMsgType.BINARY:
            return data.data

    async def iter_json(self) -> AsyncGenerator[Message, None]:
        while True:
            data = await self.receive_json()
            if data:
                yield data

    async def iter_bytes(self):
        while True:
            data = await self.receive_bytes()
            if data:
                yield data

    async def send(self, message) -> None:
        return await self.send(message)

    async def send_bytes(self, message: bytes):
        return await self.ws.send_bytes(message)

    async def send_json(self, message: Message, binary: bool = True):
        if binary:
            return await self.ws.send_bytes(bson.dumps(message))
        return await self.ws.send_json(message)

    async def close(
        self, code: Optional[int] = None, reason: Optional[str] = None
    ) -> None:
        await self.ws.close()

    def closed(self) -> bool:
        return self.ws.closed



# TODO 异常处理本地化

class WebSocketClosedError(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self) -> str:
        return super().__str__() + ": " + self.message


class JsonRpcException(Exception):
    def __init__(self, code: int, data, message) -> None:
        self.code = code
        self.data = data
        self.message = message

    def __str__(self) -> str:
        return f"{super().__str__()} {self.code} {self.data} {self.message}"





import asyncio

import orjson
import bson
import aiohttp
from aiohttp import web, WSCloseCode

from ..abcbase import BaseServer
from ...logs import logger
from ...utils import CallObj
from ...exception import WebSocketClosedError
from ...typing import Message, Optional, Tuple, List


class AioServer(BaseServer):
    def __init__(self) -> None:
        self.app = web.Application()
        self.active_connections: List[Tuple[web.WebSocketResponse, asyncio.Event]] = []

    def start(
        self,
        path: str,
        host: str,
        port: int,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:  # TODO: 更友好的终端提示
        self.loop = loop or asyncio.get_event_loop()
        self.app.add_routes([web.get(path, self.websocket_handler)])
        web.run_app(self.app, host=host, port=port, print=logger.info, loop=loop)  # type: ignore

    async def websocket_handler(self, request):
        event = asyncio.Event()
        ws = web.WebSocketResponse(autoclose=False)
        await ws.prepare(request)
        self.active_connections.append((ws, event))
        logger.success(f"{str(ws)} connected.")
        self.on_ws_callback.add_args(ws)
        await self.on_ws_callback.method(ws)
        await event.wait()
        return ws

    def disconnect(self, websocket: web.WebSocketResponse):
        for ws, event in self.active_connections:
            if ws == websocket:
                event.set()
                self.active_connections.remove((ws, event))
                break

    async def receive(self, websocket: web.WebSocketResponse):
        try:
            async for data in websocket:
                if data.type == aiohttp.WSMsgType.close:
                    raise WebSocketClosedError(data.data)
                else:
                    return data
        finally:
            if not "data" in locals():
                await self.close_ws(websocket)
                raise WebSocketClosedError(f"{self.__class__.__name__} closed.")

    async def receive_json(self, websocket: web.WebSocketResponse):
        data = await self.receive(websocket)
        if data and data.type == aiohttp.WSMsgType.TEXT:
            data = orjson.loads(data.data)
            return data
        elif data and data.type == aiohttp.WSMsgType.BINARY:
            data = bson.loads(data.data)
            return data

    async def receive_bytes(self, websocket: web.WebSocketResponse):
        data = await self.receive(websocket)
        if data and data.type == aiohttp.WSMsgType.BINARY:
            return data.data

    async def iter_bytes(self, websocket: web.WebSocketResponse):
        while True:
            data = await self.receive_bytes(websocket)
            if data:
                yield data

    async def iter_json(self, websocket: web.WebSocketResponse):
        while True:
            data = await self.receive_json(websocket)
            if data:
                yield data

    async def send_json(
        self, websocket: web.WebSocketResponse, message: dict, binary: bool = True
    ):
        if binary:
            return await websocket.send_bytes(bson.dumps(message))
        return await websocket.send_json(message)

    async def send_bytes(self, websocket: web.WebSocketResponse, message: bytes):
        return await websocket.send_bytes(message)

    async def close_ws(self, websocket: web.WebSocketResponse):
        self.disconnect(websocket)
        logger.info(f"{str(websocket)} disconnected.")

    async def close(self):
        ...

    async def on_shutdown(self):
        for ws, event in self.active_connections:
            await ws.close(code=WSCloseCode.GOING_AWAY, message=b"Server shutdown")
            event.set()

    def set_on_ws(self, func, *args, **kwargs):
        self.on_ws_callback = CallObj(method=func, args=args, kwargs=kwargs)



from pydantic import BaseModel

from ...typing import Param, Optional, Literal, Any
from ..basestandard import JsonRpcError, ProxyObj


class JsonRpc2Request(BaseModel):
    """
    JSON-RPC 2.0
    """

    jsonrpc: Literal["2.0"] = "2.0"
    id: str
    method: str
    params: Optional[Any] = None


class JsonRpc2Response(BaseModel):
    """
    JSON-RPC 2.0
    """

    jsonrpc: Literal["2.0"] = "2.0"
    id: Optional[str] = None
    result: Optional[Any] = None
    error: Optional[JsonRpcError] = None


def analysis_jsonrpc2_message(message: JsonRpc2Request):
    for i in message.method.split("."):
        yield {"method": "__getattr__", "args": (i,), "kwargs": {}}
    if isinstance(message.params, dict):
        kw = message.params
        args = ()
    elif isinstance(message.params, list):
        args = tuple(message.params)
        kw = {}
    elif message.params == None:
        return
    else:
        raise TypeError("Invalid JSON-RPC 2.0 request")
    yield {"method": "__call__", "args": args, "kwargs": kw}


def proxy_to_jsonrpc2(proxy: ProxyObj):
    """
    JSON-RPC 2.0
    """
    method = []
    params = None
    state = False
    for i in proxy:
        if state:
            raise TypeError("Invalid JSON-RPC 2.0 request")
        if isinstance(i, str):
            method.append(i)
        else:
            state = True
            params = i
    
    method = ".".join(method)

    if params == None:
        if state:
            return method, {}
        else:
            return method, params

    if params[0] and params[1]:
        raise TypeError("Invalid JSON-RPC 2.0 request")
    elif params[1]:
        params = params[1]
    elif params[0]:
        params = params[0]
    else:
        params = {}

    return method, params



from pydantic import BaseModel

from ...typing import Param, Optional, Union, List, Tuple, Literal, Any
from ..basestandard import JsonRpcError, ProxyObj, analysis_proxy


class JsonRpcXRequest(BaseModel):
    """
    JSON-RPC X
    """
    jsonrpc: Literal['X'] = 'X'
    id: str
    method: List[Union[str, Tuple[tuple, dict]]]

class JsonRpcXResponse(BaseModel):
    """
    JSON-RPC X
    """
    jsonrpc: Literal['X'] = 'X'
    id: Optional[str] = None
    result: Optional[Any] = None
    error: Optional[JsonRpcError] = None

def proxy_to_jsonrpcx(proxy: ProxyObj):
    """
    JSON-RPC X
    """
    return list(proxy)

def analysis_jsonrpcX_message(message: JsonRpcXRequest):
    return analysis_proxy(message.method)


from .basestandard import JsonRpcError, ProxyObj, Proxy, analysis_proxy, JsonRpcCode, Assign, T_Res, T_Req
from .jsonrpc2 import JsonRpc2Request, JsonRpc2Response, proxy_to_jsonrpc2, analysis_jsonrpc2_message
from .jsonrpcX import JsonRpcXRequest, JsonRpcXResponse, proxy_to_jsonrpcx, analysis_jsonrpcX_message

from ..typing import Generic, TypeVar

class Standard(Generic[T_Res, T_Req]):

    @staticmethod
    def json_to_request(json_data: dict):
        if json_data['jsonrpc'] == '2.0':
            return JsonRpc2Request(**json_data)
        if json_data['jsonrpc'] == 'X':
            return JsonRpcXRequest(**json_data)
        raise TypeError('jsonrpc must be 2.0 or X')

    @staticmethod
    def json_to_response(json_data: dict):
        if json_data['jsonrpc'] == '2.0':
            return JsonRpc2Response(**json_data)
        if json_data['jsonrpc'] == 'X':
            return JsonRpcXResponse(**json_data)
        raise TypeError('jsonrpc must be 2.0 or X')
